#ifndef lint
static char rcsid[]="$Header: gdcdaz.c,v 1.3 90/12/08 22:06:43 julian Exp $";
#endif
/*
 * Distance-azimuth routines based on geocentric direction cosines
 */
#include <math.h>
#include <local/stdtyp.h>

/* Subscripts for vector components */
#define X 0
#define Y 1
#define Z 2

/* External static variables for communication between routines */
static double	a[3];	/* Reference point				*/
static double	e[3];	/* Eastward  unit vector through point "a"	*/
static double	s[3];	/* Southward unit vector through point "a"	*/

/* Store reference point */
void
dcrefpt(g)
	double	g[3];	/* GDC of reference point	(input)		*/
{
	double	r;

	a[X] = g[X]; a[Y] = g[Y]; a[Z] = g[Z];
	if ((r = hypot(g[X], g[Y])) == 0.0) {
		/* Point is at pole */
		e[X] = e[Z] = s[Y] = s[Z] = 0.0;
		e[Y] = 1.0;
		s[X] = g[Z];
	}
	else {
		s[X] =   g[Z] *(e[Y] =   g[X] /r);
		s[Y] = (-g[Z])*(e[X] = (-g[Y])/r);
		e[Z] = 0.0;
		s[Z] = -r;
	}
}

/* Compute distance, azimuth from the reference point to a second point */
void
dcdaz(b, cosd, sind, cosz, sinz)
	double	b[3];		/* GDC of second point		(input)	*/
	double	*cosd, *sind;	/* cos(delta), sin(delta)	(output)*/
	double	*cosz, *sinz;	/* cos(az)   , sin(az)		(output)*/
{
	double	ab[3];	/* a X b */

	*cosd = a[X]*b[X] + a[Y]*b[Y] + a[Z]*b[Z];
	ab[X] = a[Y]*b[Z] - a[Z]*b[Y];
	ab[Y] = a[Z]*b[X] - a[X]*b[Z];
	ab[Z] = a[X]*b[Y] - a[Y]*b[X];
	if ((*sind = sqrt(ab[X]*ab[X] + ab[Y]*ab[Y] + ab[Z]*ab[Z])) == 0.0) {
		/* Delta = 0 or PI: Azimuth undefined - Call it zero */
		*cosz = 1.0;
		*sinz = 0.0;
	}
	else {
		*cosz = (-e[X]*ab[X] - e[Y]*ab[Y])/(*sind);
		*sinz = (b[X]*e[X] + b[Y]*e[Y])/(*sind);
	}
}

/* Compute azimuth from a second point to the reference point */
void
dcbaz(b, cosz, sinz)
	double	b[3];		/* GDC of second point		(input) */
	double	*cosz, *sinz;	/* cos(az)   , sin(az)		(output)*/
{
	double  ba[3];	/* a X b */
	double	eb[2];	/* Eastward unit vector through point "b"	*/
	double	r;
	double	sind;		/* sin(delta)				*/

	if ((r = hypot(b[X], b[Y])) == 0.0) {
		/* Point is at pole */
		eb[X] = 0.0;
		eb[Y] = 1.0;
	}
	else {
		eb[Y] =   b[X] /r;
		eb[X] = (-b[Y])/r;
	}

	ba[X] = b[Y]*a[Z] - b[Z]*a[Y];
	ba[Y] = b[Z]*a[X] - b[X]*a[Z];
	ba[Z] = b[X]*a[Y] - b[Y]*a[X];
	if ((sind = sqrt(ba[X]*ba[X] + ba[Y]*ba[Y] + ba[Z]*ba[Z])) == 0.0) {
		/* Delta = 0 or PI: Azimuth undefined - Call it zero */
		*cosz = 1.0;
		*sinz = 0.0;
	}
	else {
		*cosz = (-eb[X]*ba[X] - eb[Y]*ba[Y])/sind;
		*sinz = (a[X]*eb[X] + a[Y]*eb[Y])/sind;
	}
}

/* Determine point "b" given its distance, azimuth from reference point */
void
dcback(cosd, sind, cosz, sinz, b)
	double	cosd, sind;	/* cos(delta), sin(delta)	(input)	*/
	double	cosz, sinz;	/* cos(az)   , sin(az)		(input)	*/
	double	b[3];		/* Geocentric direction cosines	(output)*/
{
	double	t;	/* Components of unit vector tangent to great circle */
	int	i;

	for (i=0; i<3; i++) {
		t = e[i]*sinz - s[i]*cosz;
		b[i] = a[i]*cosd + t*sind;
	}
}

#ifdef TEST
/* Test driver */
#include <local/mathconst.h>
main()
{
	double	lat, lon;
	double	a[3];
	double	cosd, sind, cosz, sinz;
	double	cbz, sbz;
	double	junk;
	
	scanf("%lf%lf", &lat, &lon);
	cvscd(1.0, RAD*(90.0-lat), RAD*lon, &a[X], &a[Y], &a[Z]);
	dcrefpt(a);
	while (2 == scanf("%lf%lf", &lat, &lon)) {
		cvscd(1.0, RAD*(90.0-lat), RAD*lon, &a[X], &a[Y], &a[Z]);
		dcdaz(a, &cosd, &sind, &cosz, &sinz);
		printf("delta = %g, az = %g",
			DEG*atan2(sind, cosd), DEG*atan2(sinz, cosz));
		dcbaz(a, &cbz, &sbz);
		printf(", back az = %g\n", DEG*atan2(sbz, cbz));
		dcback(cosd, sind, cosz, sinz, a);
		cvcsd(a[X], a[Y], a[Z], &junk, &lat, &lon);
		lat = HALFPI - lat;
		printf("lat = %g, lon = %g\n", DEG*lat, DEG*lon);
	}
}
#endif
