#ifndef lint
static char rcsid[]="$Header: jacobi.c,v 1.1 88/10/17 13:58:07 julian Exp $";
#endif
/*
 * Eigenvalues, eigenvectors of real symmetric matrix
 * From sec. 11.1 of
 * Press, Flannery, Teukolsky, and Vetterling,
 * Numerical Recipies in C, Cambridge Univ. Press, 1988.
 */

#include <math.h>

#define ROTATE(a, i, j, k, l)	{ g=a[i][j]; h=a[k][l]; \
				a[i][j]=g-s*(h+g*tau); a[k][l]=h+s*(g-h*tau); }

float	*fvector();
void	free_fvector();

void
jacobi(a, n, d, v, nrot)
	float	**a;	/* Real symmetric matrix[1..n][1..n]	(in)	*/
			/* On return, elements above diagonal		*/
			/* are destroyed.				*/
	int	n;	/* Order of matrix a			(in)	*/
	float	d[];	/* Eigenvalues [1..n]			(out)	*/
	float	**v;	/* Normalized eigenvectors [1..n][1..n]	(out)	*/
			/* N.B.:  Stored columnwise			*/
	int	*nrot;	/* No. of Jacobi rotations		(out)	*/
{
	float	*b;
	float	c;
	float	g;
	float	h;
	int	i, j;	/* Dummy loop indices				*/
	int	p, q;	/* Dummy loop indices				*/
	float	s;
	float	sm;	/* Sum of abs of off-diagonal elements		*/
	float	t;
	float	tau;
	float	theta;
	float	thresh;	/* Convergence threshold			*/
	float	*z;	/* Sum of terms of the form t*a[p][q],		*/
			/* as in equation 11.1.14			*/

	b = fvector(1, n);
	z = fvector(1, n);

	/* Initialize v[][] to the identity matrix */
	for (p=1; p<=n; p++) {
		for (q=1; q<=n; q++)
			v[p][q] = 0.0;
		v[p][p] = 1.0;
	}

	/* Initialize b and d to diagonal of a */
	for (p=1; p<=n; p++) {
		b[p] = d[p] = a[p][p];
		z[p] = 0.0;
	}

	/* Iteration loop */
	*nrot = 0;
	for (i=1; i<=50; i++) {

		/* Sum off-diagonal elements */
		sm = 0.0;
		for (p=1; p<=n-1; p++)
			for (q=p+1; q<=n; q++)
				sm += fabs(a[p][q]);
		if (sm == 0.0) {
			/* Normal return */
			/* Relies on convergence to machine underflow */
			free_fvector(z, 1, n);
			free_fvector(b, 1, n);
			return;
		}

		thresh = (i<4 ? 0.2*sm/(n*n) : 0.0);

		for (p=1; p<=n-1; p++) {
			for (q=p+1; q<=n; q++) {
				g = 100.0*fabs(a[p][q]);

				/* After four sweeps, skip rotation */
				/* if off-diagonal element is small */
				if (i>4 && fabs(d[p])+g == fabs(d[p])
				    && fabs(d[q])+g == fabs(d[q]))
					a[p][q] = 0.0;
				else if (fabs(a[p][q]) > thresh) {
					h = d[q] - d[p];
					if (fabs(h)+g == fabs(h))
						/* t = 0.5/theta */
						t = a[p][q]/h;
					else {
						/* Equation 11.1.10 */
						theta = 0.5*h/a[p][q];
						t = 1.0/(fabs(theta) +
							sqrt(1.0+theta*theta));
						if (theta < 0.0)
							t = -t;
					}
					c = 1.0/sqrt(1.0+t*t);
					s = t*c;
					tau = s/(1.0+c);
					h = t*a[p][q];
					z[p] -= h; z[q] += h;
					d[p] -= h; d[q] += h;
					a[p][q] = 0.0;

					/* Case of rotations 1 <= j < p */
					for (j=1; j<p; j++)
						ROTATE(a, j, p, j, q);

					/* Case of rotations p < j < q */
					for (j=p+1; j<q; j++)
						ROTATE(a, p, j, j, q);

					/* Case of rotations q < j <= n */
					for (j=q+1; j<=n; j++)
						ROTATE(a, p, j, q, j);

					for (j=1; j<=n; j++)
						ROTATE(v, j, p, j, q);
					(*nrot)++;
				}
			}
		}

		/* Update d with sum of t*a[p][q], reinitialize z */
		for (p=1; p<=n; p++) {
			b[p] += z[p];
			d[p] = b[p];
			z[p] = 0.0;
		}
	}
}
