#ifndef lint
static char rcsid[]="$Header: qselect.c,v 1.13 92/08/13 19:04:44 julian Exp $";
#endif lint
/*
 * Select event "summary cards" matching specified criteria
 * Bruce R. Julian, USGS Menlo Park, CA
 * Significantly altered by Doug Neuhauser, Seismographic Station, UCB
 */

#include <stdio.h>
#include <math.h>
#include <ctype.h>
#include <string.h>
#include <sys/param.h>
#include <local/local.h>
#include <local/date_time.h>
#include <local/cmd_opt.h>
#include <local/mathconst.h>
#include <local/earthconst.h>
#include <local/decode.h>

/* Types of input formats. */
#define	NCSN_CATALOG	1
#define	NCSN_MECH	2
#define UCB_CATALOG	4
#define CNSS_CATALOG	8
#define	REDI_CATALOG	16
#define FULL_CNSS_CATALOG      32

#define		ALLOC_EXTERN			/* Allocate externals.	*/
#include "global.h"
#include "procs.h"

extern int search_ucb_catalog();
extern int search_ncsn_catalog();
extern int search_ncsn_mech();
extern int search_cnss_catalog();
extern int search_redi_catalog();
extern int search_full_cnss_catalog();

extern FILE *ecfopen();

/*  Magnitude searches will be done on based on a preference list, 	*/
/*  since there are numerous magnitudes and not all events have values	*/
/*  for all of the magnitudes.						*/

/* Table that defines magtype strings and corresponding values.		*/
struct _magtypes { 
    char *name; 
    int	val;
} magtypes[] = { 
    "ml",	ML_MAG,
    "mml",	MML_MAG,
    "coda",	CODA_MAG,
    "amp",	AMP_MAG,
    "avg",	AVG_MAG,
    "mln",	MLN_MAG,
    "bm",	BM_MAG,
    "mw",	MW_MAG,
    NULL,	0 };

/*  Default preference order for each catalog type.			*/
static MAGS	ucb_catalog_mags = {	/* dflt mag search: UCB_CATALOG	*/
    3, MLN_MAG, ML_MAG, BM_MAG, MW_MAG, 0, 0, 0, 0 };
static MAGS	ncsn_mech_mags = {	/* dflt mag search: NCSN_MECH*/
    1, AVG_MAG, 0, 0, 0, 0, 0, 0, 0 };
static MAGS	ncsn_catalog_mags = {	/* dflt mag search; NCSN_CATALOG */
    4, MML_MAG, ML_MAG, CODA_MAG, AMP_MAG, 0, 0, 0, 0 };


/**/
/*----  C O M M A N D - L I N E   A R G U M E N T   P R O C E S S I N G  ----*/
dcheck(mn, mx, s)	/* Be sure mn <= mx */
	double mn, mx;
	char *s;
{
	if (mn > mx) error("minimum %s value greater than maximum", s);
}

icheck(mn,mx,s) 	/* same as dcheck -- but for integers */
int mn,mx;
char *s;
{
	if (mn > mx) error("minimum %s value greater than maximum", s);
}

/* format_work -- decode the format given on the command line, if any */
format_work(format)
char *format;
{
    if (input_format) {
	fprintf (stderr, "error - only 1 format allowed\n");
	exit(1);
    }
    else if (!strcasecmp(format,"ucb")) input_format = UCB_CATALOG;
    else if (!strcasecmp(format,"mech")) input_format = NCSN_MECH;
    else if (!strcasecmp(format, "ncsn")) input_format = NCSN_CATALOG;
    else if (!strcasecmp(format, "calnet")) input_format = NCSN_CATALOG;
    else if (!strcasecmp(format, "cnss")) input_format = CNSS_CATALOG;
    else if (!strcasecmp(format, "f_cnss")) input_format = FULL_CNSS_CATALOG;
    else if (!strcasecmp(format, "redi")) input_format = REDI_CATALOG;
    else {
	fprintf (stderr, "error - invalid format\n");
	exit(1);
    }
}

/* Convert string to angle */
/* "degrees[:minutes[:seconds]]" */
double angle(p)
char	*p;
{
	double	sign;
	double	d, m, s;

	sign = 1.0;
	if (*p == '+')
		p++;
	if (*p == '-') {
		sign = -1.0;
		p++;
	}
	d = atof(p);
	m = s = 0.;
	if ((p = index(p, ':')) != NULL) {
		m = atof(++p);
		if ((p = index(p, ':')) != NULL)
			s = atof(++p);
	}
	return sign*(d + (m + s/60.)/60.);
}

/* Check whether string looks like an angle */
bool
isangle(p)
	char	*p;
{
	if (p == (char *)NULL)
		return FALSE;
	while (*p)
		if (index("+-0123456789.:", *p++) == (char *)NULL)
			return FALSE;
	return TRUE;
}

/* Convert string to TIME */
/* "[cc]yymmddhhmmss.ss...[+-offset]" */
TIME epoch(s)
	char	*s;
{
	double	se;
	int	yr, mo, da, hr, mn;
	char	*p;
	double	offset();

	/* Get year (twentieth century default) */
	for (p=s; isdigit(*p); p++)
		;
	switch (p - s) {
	case 14:
		yr = DECODE(s, 4, atoi);
		s += 2;
		break;
	case 12:
		yr = 1900 + DECODE(s, 2, atoi);
		break;
	default:
		error ("Bad date format: \"%s\"", s);
	}

	/* Get month, day, hour, minute, second */
	mo = DECODE(s+ 2, 2, atoi);
	da = DECODE(s+ 4, 2, atoi);
	hr = DECODE(s+ 6, 2, atoi);
	mn = DECODE(s+ 8, 2, atoi);
	se = atof(s+10);

	/* Skip to end of seconds field */
	s += 12;
	if (*s == '.')
		s++;
	while(isdigit(*s))
		s++;

	return timvar(yr, mo, da, hr, mn, se, GREGORIAN) + offset(s);
}

static double offset(s)
	char	*s;
{
	double	sign, x, offset;

	switch (*s) {
	case '\0':	return 0.;
	case '+':	sign = 1.;	break;
	case '-':	sign = -1.;	break;
	default:
		error("Illegal sign character \"%c\"", *s);
	}
	s++;
	offset = 0.;
	while (isdigit(*s) || *s == '.') {
		x = atof(s);
		while (isdigit(*s))	s++;
		if (*s == '.')		s++;
		while (isdigit(*s))	s++;
		switch (*s) {
		case 'd': case 'D': x *= DAY	; break;
		case 'h': case 'H': x *= HOUR	; break;
		case 'm': case 'M': x *= MINUTE	; break;
		case 's': case 'S': x *= SECOND	; break;
		default:
			error("Bad time offset unit \"%c\"", *s);
		}
		s++;
		offset += x;
	}
	return sign*offset;
}

/* Smaller of two double values */
double fmin2(a, b)
	double	a, b;
{
	return (a<b ? a : b);
}

/* Larger of two double values */
double fmax2(a, b)
	double	a, b;
{
	return (a>b ? a : b);
}

qhelp	(){ if(eargc>2) prt_help(); else prt_doc("[file...]"); }
qneg    (){ negate = TRUE; }
qlat    (){ latmin = angle(aarg()); latmax = angle(aarg()); dcheck(latmin, latmax, "latitude"); do_lat = TRUE; }
qlon    (){ lonmin = angle(aarg()); lonmax = angle(aarg()); dcheck(lonmin, lonmax, "longitude"); do_lon = TRUE; }
qdepth  (){ zmin   = narg(); zmax = narg(); dcheck(zmin, zmax, "depth"); do_depth = TRUE; }
qtime   (){ otmin  = epoch(aarg()); otmax = epoch(aarg()); dcheck(otmin, otmax, "time"); do_otime = TRUE; }
qmtype  (){ parsemag(aarg()); }
qmag    (){ magmin = narg(); magmax = narg(); dcheck(magmin, magmax, "magnitude"); do_mag = TRUE; }
qno_mag (){ do_nomag=TRUE;}
qclose  (){ clomin = narg(); clomax = narg(); dcheck(clomin, clomax, "close"); do_close = TRUE; }
qnsta   (){ nstmin = narg(); nstmax = narg(); icheck(nstmin, nstmax, "nstations"); do_nsta = TRUE; }
qrms    (){ rmsmin = narg(); rmsmax = narg(); dcheck(rmsmin, rmsmax, "rms"); do_rms = TRUE; }
qgap    (){ gapmin = narg(); gapmax = narg(); dcheck(gapmin, gapmax, "gap"); do_gap = TRUE; }
qcat    (){ search_catalog = TRUE; }
qverbose(){ verbose = TRUE; }
qformat (){ format_work(aarg()); }
qnmagobs(){ nmagmin = narg(); nmagmax = narg(); icheck(nmagmin, nmagmax, "nobs for mag"); do_mag_obs = TRUE; }
qm0	(){ m0min = narg(); m0max = narg(); dcheck(m0min, m0max, "M0"); do_m0 = TRUE; }
qnm0obs (){ nm0min = narg(); nm0max = narg(); icheck(nm0min, nm0max, "nobs for M0"); do_m0_obs = TRUE; }
qoterr  (){ otemin = narg(); otemax = narg(); dcheck(otemin, otemax, "o-time error"); do_oterr = TRUE; }
qlaterr (){ laemin = narg(); laemax = narg(); dcheck(laemin, laemax, "lat error"); do_laterr = TRUE; }
qlonerr (){ loemin = narg(); loemax = narg(); dcheck(loemin, loemax, "lon error"); do_lonerr = TRUE; }
qherr	(){ herrmin = narg(); herrmax = narg();	dcheck(herrmin, herrmax, "horizontal error"); do_horerr = TRUE; }
qzerr	(){ zerrmin = narg(); zerrmax = narg();	dcheck(zerrmin, zerrmax, "depth error"); do_deperr = TRUE; }
qfelt	(){ felt = 'F'; do_felt = TRUE; }
qqual   (){ char *p; quals=aarg(); for(p=quals;*p;p++) *p = (isupper(*p) ? tolower(*p) : *p); do_qual = TRUE; }
qetype	(){ char *p; etypes=aarg(); for(p=etypes;*p;p++) *p = (islower(*p) ? toupper(*p) : *p); do_etype = (strchr(etypes,'A')) ? FALSE : TRUE; }
qdelta  ()
{
	double	rplat;
	double	dlon, sdlon;

	rplat  = RAD*angle(aarg());
	phi0   = RAD*angle(aarg());
	delmin = narg()/REARTH;
	delmax = narg()/REARTH;
	dcheck(delmin, delmax, "delta");

	/* Do all possible preliminary arithmetic */
	rplat = ggtogc(rplat);
	st0 = cos(rplat);
	ct0 = sin(rplat);
	cdmin = cos(delmin);
	cdmax = cos(delmax);

	/* Shrink lat-lon limits to barely enclose delta annulus */
	if (! do_lat) { latmin =  -90.; latmax =  90.; }
	if (! do_lon) { lonmin = -180.; lonmax = 180.; }
	latmin = fmax2(latmin, DEG*gctogg(rplat-delmax));
	latmax = fmin2(latmax, DEG*gctogg(rplat+delmax));
	if (st0 > 0.) {
		sdlon = sin(delmax)/st0;
		if (sdlon <= 1.) {
			dlon = asin(sdlon);
			lonmin = fmax2(lonmin, DEG*(phi0-dlon));
			lonmax = fmin2(lonmax, DEG*(phi0+dlon));
		}
	}
	do_delta = do_lat = do_lon = TRUE;
}

qpoly   ()
{
	/* Read in vertices of polygon */
	for (nvert=0; isangle(peekarg()) && nvert<MAXVERT; nvert++) {
		plat[nvert] = angle(aarg());
		plon[nvert] = angle(aarg());
	}
	if (nvert < 3)
		error("Polygon needs at least three vertices");
	plat[nvert] = plat[0];
	plon[nvert] = plon[0];
	nvert++;

	/* Shrink latitude-longitude limits to just include polygon */
	if (! do_lat) { latmin =  -90.; latmax =  90.; }
	if (! do_lon) { lonmin = -180.; lonmax = 180.; }
	latmin = fmax2(latmin, dmin(plat, nvert));
	latmax = fmin2(latmax, dmax(plat, nvert));
	lonmin = fmax2(lonmin, dmin(plon, nvert));
	lonmax = fmin2(lonmax, dmax(plon, nvert));
	do_poly = do_lat = do_lon = TRUE;
}

struct command cmd[] = {
   qhelp    , "-help"     , "[opt...]" , "Print documentation",
   qneg     , "-negate"   , ""         , "Print events that FAIL test",
   qtime    , "-time"     , "min max"  , "Origin-time interval ([cc]yymmddhhmmss[+-offset])",
   qlat     , "-latitude" , "min max"  , "Latitude range (d[:m[:s]])",
   qlon     , "-longitude", "min max"  , "Longitude range (d[:m[:s]]) N.B. West is negative!",
   qdepth   , "-depth"    , "min max"  , "Depth range (km)",
   qmtype   , "-mtype"    , "type,..." , "Magnitude type[s], in preferred order: (coda, amp, avg, ml, mml, mln, bm, mw)",
   qmag     , "-magnitude", "min max"  , "Magnitude range",
   qclose   , "-closest"  , "min max"  , "Range of closest station (km)",
   qnsta    , "-nstations", "min max"  , "Range of no. of stations",
   qrms     , "-rms"      , "min max"  , "Range of r.m.s. residual (s)",
   qgap     , "-gap"      , "min max"  , "Range of azimuthal gap (deg)",
   qdelta   , "-delta"    , "lat lon min max"  , "Epicentral distance (km)",
   qpoly    , "-polygon"  , "lat lon...","Polygon (give 3 or more vertices)",
   qcat     , "-catalog"  , ""         , "Search catalog - uses CUSPCATDIR, CUSPMECHDIR, UCBCATDIR, or CNSSCATDIR variable",
   qverbose , "-verbose"  , ""         , "Verbose output during search",
   qformat  , "-format"   , "type"     , "Specify input data format: (ncsn, ucb, mech, cnss, or redi)",
   qqual    , "-quality"  , "letters"  , "Quality codes (e.g. abc)   [ucb only]",
   qetype   , "-etype"	  , "letters"  , "Event types (eg eqnrfa) [ncsn, cnss only]",
   qnmagobs , "-nmag"     , "min max"  , "Range of number of observations for magnitude   [ucb only]",
   qm0      , "-m0"	  , "min max"  , "Range of scalar moment   [ucb only]",
   qnm0obs  , "-nm0    "  , "min max"  , "Range of number of observations for scalar moment   [ucb only]",
   qoterr   , "-errtime"  , "min max"  , "Range of error in origin time (sec)   [ucb only]",
   qlaterr  , "-errlat"   , "min max"  , "Range of error in latitude (km)   [ucb only]",
   qlonerr  , "-errlon"   , "min max"  , "Range of error in longitude (km)   [ucb only]",
   qherr    , "-errhoriz" , "min max"  , "Range of error in horizontal location (km)   [ncsn, mech, cnss only]",
   qzerr    , "-errdepth" , "min max"  , "Range of error in focal depth (km)",
   qfelt    , "-felt"     , ""         , "Select felt events   [ucb cnss ncsn only]",
   qno_mag  , "-no_mag"   , ""         , "Include events which have no reported magnitude [cnss]",
};

/*--------------------------  M A I N  ------------------------------*/

main(argc,argv)
	int	argc;
	char	*argv[];
{
	FILE *fpi_cat;
	char filename[MAXPATHLEN],*catdir;
	char *catalog_name;
	int (*search_f)();		/* ptr to func returning int.	*/
	char *location;			/* location of files.		*/

	/* Set defaults */
	negate = FALSE;
	search_catalog = verbose = FALSE;
	do_lat = do_lon = do_depth = do_otime = do_mag = 
	do_close = do_nsta = do_rms = do_gap = do_qual =
	do_delta = do_poly = do_nomag =FALSE;		/* Do nothing		*/

	do_etype = TRUE; etypes = "E";

	/* Interpret command-line options */
	PROCESS_OPTS;

	/* Establish defaults if not explicitly set. */
	if (!input_format) input_format = NCSN_CATALOG;
	if ((do_mag || do_nomag)&& mags.nprefs == 0) {
	    set_def_mags(input_format);
	}

	/* Process files (standard input default) */
	switch (input_format) {
	  case NCSN_CATALOG:
	    location = "CUSPCATDIR";
	    search_f = search_ncsn_catalog;
	    break;
	  case NCSN_MECH:
	    location = "CUSPMECHDIR";
	    search_f = search_ncsn_mech;
	    break;
	  case UCB_CATALOG:
	    location = "UCBCATDIR";
	    search_f = search_ucb_catalog;
	    break;
	  case CNSS_CATALOG:
	    location = "CNSSCATDIR";
	    search_f = search_cnss_catalog;
	    break;
	  case FULL_CNSS_CATALOG:
	    location = "FULLCNSSCATDIR";
	    search_f = search_full_cnss_catalog;
	    break;
	  case REDI_CATALOG:
	    location = "REDICATDIR";
	    search_f = search_redi_catalog;
	    break;
	  default:
	    fprintf (stderr, "unrecognized input format - should never get here\n");
	    exit(1);
	}

	if ((iarg >= argc) && ! search_catalog) {
	    if (verbose) fprintf (stderr, "processing stdin ...\n");
	    (*search_f)(stdin);
	}

	else {
	    if (search_catalog) {
		TIME catmin = otmin;
		TIME catmax = otmax;
		char **cp, **find_catalog();

		if ((cp = find_catalog(location,catmin,catmax))==NULL) {
		    fprintf(stderr,"unable to find installed files for specified input format.\n");
		    exit(1);
		}
		while (*cp != NULL) {
		    FILE *f = ecfopen(*cp, "r");
		    if (verbose) fprintf(stderr,"processing %s ... ", *cp);
		    (*search_f)(f);
		    (void)ecfclose(f);
		    cp++;
		}
	    }
	    for (; iarg<argc; iarg++) {
		FILE *f = ecfopen(argv[iarg], "r");
		if (verbose) fprintf(stderr,"processing %s ... ", argv[iarg]);
		(*search_f)(f);
		(void)ecfclose(f);
	    }
	}
	exit(0);
}

set_def_mags(format)
     int format;
{
    switch (format) {
      case NCSN_CATALOG:
	bcopy (&ncsn_catalog_mags, &mags, sizeof (mags));
	break;
      case NCSN_MECH:
	bcopy (&ncsn_mech_mags, &mags, sizeof (mags));
	break;
      case UCB_CATALOG:
	bcopy (&ucb_catalog_mags, &mags, sizeof (mags));
	break;
      case CNSS_CATALOG:
	bcopy ("", &mags, 1);
	break;
      case FULL_CNSS_CATALOG:
	bcopy ("", &mags, 1);
	break;
      case REDI_CATALOG:
	bcopy ("", &mags, 1);
	break;
      default:
	fprintf (stderr, "unknown magnitude pref list for this format\n");
	exit(1);
	break;
    }
}

/* Parse magnitude string into structure. */
int
parsemag(p)
     char *p;
{
    char *q;
    int i;

    while (p != NULL && *p !=  '\0' && mags.nprefs < MAXMAGTYPES) {
	q = strchr (p, ',');
	if (q != NULL) *q = '\0';
	for (i=0; i < MAXMAGTYPES; i++) {
	    if (strcasecmp(p, magtypes[i].name)==0) {
		mags.magtype[mags.nprefs] = magtypes[i].val;
		++mags.nprefs;
		break;
	    }
	}
	if (i >= MAXMAGTYPES) {
	    fprintf (stderr, "invalid magtype: %s\n", p);
	    exit(1);
	}
	p = (q) ? q+1 : q;
    }
    return(1);
}
