#ifndef lint
static char rcsid[] = "$Header: qloc.c,v 1.4 92/05/25 07:28:59 julian Exp $";
#endif lint
/*
 * Command to estimate origins of seismic events
 * Bruce R. Julian, USGS Menlo Park, Calif.
 */

#include <stdio.h>
#include <errno.h>
#include <math.h>
#include <local/cmd_opt.h>
#include <local/mathconst.h>
#include <local/quake.h>
#include <local/bwflat.h>

#define CHUNKSIZE 50		/* No. of ptrs allocated each time	*/
#define DEFMDL	"/we/crhet/julian/quake/src/cmd/qloc/test_dir/model1"

/* Initialize control structure to default values */
RLCTRL		ctl = {			/* Control structure		*/
	{FALSE, FALSE, FALSE, FALSE},	/* Nothing restrained		*/
	2e-6, 2e-6, 0.01,		/* Convergence thresholds 	*/
	0.0, 1000.0,			/* Focal-depth limits		*/
	10,				/* Iteration limit		*/
	5,				/* Damping limit	 	*/
	100.0,				/* Starting damping parameter	*/
	50.0,				/* Minimum damping parameter	*/
	5.0,				/* Multiplier per damp step	*/
	2.0,				/* Relative depth damping	*/
	0.02,				/* mu				*/
	NEVER,				/* print flag			*/
};
/**/
/*                 E X T E R N A L   V A R I A B L E S                  */

/*	Error handling and reporting					*/
int		errno;			/* Error flag			*/

/*	Storage for event origin and statistics pertaining thereto	*/
ORIGIN		origin;			/* Origin structure		*/
STATIS		o_stat;			/* Statistics structure		*/

/*	Readings and related info.					*/
/*	(Dynamically allocated pointer arrays)				*/
BWRSLT		**bwrslt;		/* Body-wave result structures	*/
READING		**rdgs;			/* Reading structures		*/
ASSOC		**assoc;		/* Association structures	*/
int		nrdg;			/* Number of readings		*/

/*	Variables set by optional command-line arguments		*/
FILE		*fig;			/* File of stations to ignore	*/
FILE		*fmdl;			/* Model file			*/
FILE		*fprt;			/* Output print file		*/
tbool		guess;			/* Make initial origin estimate */
tbool		setdepth;		/* Did user set focal depth?	*/
tbool		setlat, setlon;		/* Did user set coordinates?	*/
tbool		settime;		/* Did user set origin time?	*/
float		vdepth;			/* User-supplied focal depth	*/
float		vlat, vlon;		/* User-supplied coordinates	*/
TIME		vtime;			/* User-supplied origin time	*/

/* Functions */
void		bytcpy();
void		cvcsd();
FILE		*efopen();
char		*emalloc();
void		error();
char		*realloc();
char		*ermsg();
int		reloc();
char		*strcat();
char		*strcpy();

/**/
/* Get angle from degree and optional minute arguments */
static double getdm()
{
	double	d=narg(), m=0.0;
	if (isnarg() && d*(m = narg()) < 0.0)
		m = (-m);
	return RAD*(d + m/60.0);
}

/*-----------ARGUMENT-PROCESSING ROUTINES----------------*/
static fhelp   (){ if (eargc>2) prt_help(); else prt_doc("< data_file"); }
static fdamp   (){ ctl.lstart = narg(); ctl.lmin = narg(); ctl.lfact = narg(); ctl.max_damp = (int)narg(); }
static ffixlat (){ ctl.hold[ILAT] = TRUE; if (isnarg()) fsetlat(); }
static ffixlon (){ ctl.hold[ILON] = TRUE; if (isnarg()) fsetlon(); }
static ffixt   (){ ctl.hold[ITIME] = TRUE; }
static ffixdep (){ ctl.hold[IDEPTH] = TRUE; if (isnarg()) fsetdep(); }
static fguess  (){ guess = TRUE; }
static fignore (){ fig = efopen(aarg(), "r"); }
static fmaxdep (){ ctl.max_depth = narg(); }
static fmaxit  (){ ctl.max_iter = (int)narg(); }
static fmindep (){ ctl.min_depth = narg(); }
static fmodel  (){ fmdl = efopen(aarg(), "r"); }
static fmu     (){ ctl.mu = narg(); }
static fnoprint(){ ctl.iprint = NEVER; }
static fprntall(){ ctl.iprint = ALL; fprt = efopen(aarg(), "w"); }
static fprntlas(){ ctl.iprint = FINAL; fprt = efopen(aarg(), "w"); }
static fsetlat (){ setlat = TRUE; vlat = getdm(); }
static fsetlon (){ setlon = TRUE; vlon = getdm(); }
static fsetdep (){ setdepth = TRUE; vdepth = narg(); }
static funbuf  (){ setbuf(stdout,(char *)NULL); if (fprt) setbuf(fprt,(char *)NULL); }
static fzdamp  (){ ctl.zdamp = narg(); }

/*----------DRIVING TABLE FOR ARGUMENT PROCESSING----------*/
struct command cmd[] = {
   fhelp    , "-help"     , "[opt...]"      , "Documentation help",
   fdamp    , "-damp"     , "L0 Lmin Lfact n", "Set damping parameters",
   ffixlat  , "-fixlat"   , "[deg [min]]"   , "Restrain [and set] latitude",
   ffixlon  , "-fixlon"   , "[deg [min]]"   , "Restrain [and set] longitude",
   ffixt    , "-fixtime"  , ""              , "Restrain origin time",
   ffixdep  , "-fixdep"   , "[km]"          , "Restrain [and set] focal depth",
   fguess   , "-guess"    , ""              , "Make initial estimate of origin",
   fignore  , "-ignore"   , "filename"      , "File of stations to ignore",
   fmaxdep  , "-maxdepth" , "km"            , "Maximum allowed focal depth",
   fmaxit   , "-maxiter"  , "n"             , "Max. no. of iterations",
   fmindep  , "-mindepth" , "km"            , "Minimum allowed focal depth",
   fmodel   , "-model"    , "filename"      , "Model file",
   fmu      , "-mu"       , "value"         , "Frequency of blunders in readings",
   fnoprint , "-noprint"  , ""              , "Suppress printed output",
   fprntall , "-printall" , "filename"      , "Print all iterations to file",
   fprntlas , "-printlast", "filename"      , "Print final iteration to file",
   fsetlat  , "-setlat"   , "deg [min]"     , "Set initial latitude",
   fsetlon  , "-setlon"   , "deg [min]"     , "Set initial longitude",
   fsetdep  , "-setdep"   , "km"            , "Set initial focal depth",
   funbuf   , "-unbuff"   , ""              , "Unbuffered output",
   fzdamp   , "-depdamp"  , "value"         , "Set depth damping",
};
/**/
/*--------------------------  M A I N  ------------------------------*/

main(argc,argv)
	int	argc;
	char	*argv[];
{
	void	do_guess();
	void	init();
	int	getdat();
	MODEL	model;			/* Earth model			*/
	void	putdat();
	int	i;			/* Loop index			*/
	int	j;			/* Loop index			*/
	double	zmax;			/* Total thickness of submodel	*/

	/* Process arguments */
	init();
	PROCESS_OPTS

	/* Read in earth model */
	if (fmdl == NULL)
		fmdl = efopen(DEFMDL, "r");
	getmdl(fmdl, &model);
	(void)fclose(fmdl);

	/* Insure that events remain inside earth model */
	for (i=0; i<model.nsm; i++) {
		zmax = 0.0;
		for (j=0; j<model.sm[i].nl; j++)
			zmax += model.sm[i].h[j];
		ctl.max_depth = MIN(zmax*(1.0-100.0*EPS), ctl.max_depth);
	}

	/* Event loop */
	while ((nrdg=getdat(stdin, &rdgs, &assoc, &origin)) > 0) {
		if (guess)
			do_guess(rdgs, nrdg, &origin);
		if (setlat)
			origin.o_lat = vlat;
		if (setlon)
			origin.o_lon = vlon;
		if (setdepth)
			origin.o_depth = vdepth;
		if (settime)
			origin.o_time = vtime;

		/* Allocate body-wave result array */
		bwrslt = (BWRSLT **)emalloc((unsigned)nrdg*sizeof(BWRSLT *));
		for (i=0; i<nrdg; i++)
			bwrslt[i] = (BWRSLT *)emalloc((unsigned)sizeof(BWRSLT));

		if (reloc(fprt, &model, &origin, rdgs, bwave,
			  assoc, bwrslt, nrdg, &ctl, &o_stat) != 0)
			(void)fprintf(stderr, "%s: %s\n", progname, ermsg(errno));
		putdat(stdout, nrdg, rdgs, assoc, &origin);
		for (i=0; i<nrdg; i++) {
			free((char *)rdgs[i]);
			free((char *)assoc[i]);
			free((char *)bwrslt[i]);
		}
		free((char *)rdgs);
		free((char *)assoc);
		free((char *)bwrslt);
	}
	exit(0);
}
/**/
static void
init()
{
	fprt = fmdl = fig = NULL;
	guess	= FALSE;
	setlat = setlon = setdepth = settime = FALSE;
}

/* Make initial guess of origin */
static void
do_guess(rp, n, op)
	READING			*rp[];
	int			n;
	ORIGIN			*op;
{
	TIME	early = HUGE;
	READING	*erp;		/* Earliest reading */
	double	junk, theta, phi;

	while (n-- > 0) {
		if ((~(*rp)->flags & (ASSOCIATED | T_EXISTS | USE_T)) == 0 && (*rp)->atime < early) {
			erp = *rp;
			early = erp->atime;
		}
		rp++;
	}
	op->o_time  = early;
	cvcsd(erp->sta.gdc[X], erp->sta.gdc[Y], erp->sta.gdc[Z], &junk, &theta, &phi);
	op->o_lat   = HALFPI - theta;
	op->o_lon   = phi;
	op->o_depth = 10.0;
}
/**/
/* Read observations, origin	*/
static int
getdat(f, rdg_a, assoc_a, optr)
	FILE		*f;		/* Input unit		(input)	*/
	READING		*(*rdg_a[]);	/* Array of reading ptrs(output)*/
	ASSOC		*(*assoc_a[]);	/* Array of assoc ptrs	(output)*/
	ORIGIN		*optr;		/* Origin struct	(output)*/
{
	int		n;
	READING		r_buff;
	ASSOC		a_buff;
	unsigned	asz;

	if (1 != fread((char*)optr, sizeof(ORIGIN), 1, f))
		return FAIL;
	asz = CHUNKSIZE;
	*rdg_a = (READING **)emalloc(asz*sizeof(READING *));
	*assoc_a = (ASSOC **)emalloc(asz*sizeof(ASSOC *));
	n = 0;
	while (1 == fread((char*)&r_buff, sizeof(r_buff), 1, f) &&
	       1 == fread((char*)&a_buff, sizeof(a_buff), 1, f)) {
		if (*r_buff.sta.s_code == '\0')
			break;
		if (n >= asz) {
			asz += CHUNKSIZE;
			*rdg_a   = (READING **)realloc((char *)*rdg_a,   asz*sizeof(READING *));
			*assoc_a = (ASSOC   **)realloc((char *)*assoc_a, asz*sizeof(ASSOC *));
			if (*rdg_a == NULL || *assoc_a == NULL)
				error("Can't realloc");
		}
		(*rdg_a)[n] = (READING *)emalloc((unsigned)sizeof(READING));
		bytcpy((char *)&r_buff, (char *)(*rdg_a)[n], sizeof(READING));
		(*assoc_a)[n] = (ASSOC *)emalloc((unsigned)sizeof(ASSOC));
		bytcpy((char *)&a_buff, (char *)(*assoc_a)[n], sizeof(ASSOC));
		n++;
	}
	if (n <= 0)
		error("No readings in input");
	return n;
}

static void
putdat(f, n, rptr, aptr, optr)
	FILE		*f;		/* Output unit		(input)	*/
	int		n;		/* No. of readings	(input) */
	READING		*rptr[];	/* Reading-ptr array	(input)	*/
	ASSOC		*aptr[];	/* assoc-ptr array	(input) */
	ORIGIN		*optr;		/* Origin struct	(input)	*/
{
	int		i;

	if (1 != fwrite((char*)optr, sizeof(ORIGIN), 1, f))
		error("Can't write out readings");
	while (n--) {
		if (1 != fwrite((char*)(*rptr++), sizeof(READING), 1, f) ||
		    1 != fwrite((char*)(*aptr++), sizeof(ASSOC),   1, f))
			error("Can't write out readings");
	}

	/* Write out empty reading, assoc structures */
	i = sizeof(READING) + sizeof(ASSOC);
	while (i--)
		if (putc('\0', f) == EOF)
			error("Can't write out delimiter");
}
